#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mfhdf.h"

#include "const.h"
#include "types.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "cloud.h"
#include "bowtie.h"
#include "confidence.h"
#include "frp.h"
#include "zstats.h"
#include "cmg.h"


#define BIT_MASK(x) (1U << x)


#undef max
#define max(a,b) (a < b ? b : a)


#define MAX_WIN_SIZE 21
#define MAX_WIN_PIX (MAX_WIN_SIZE * MAX_WIN_SIZE)


typedef struct {
	/* thresholds used to identify potential fire pixels */
	float	thresh_PF_T21;	/* K */
	float	thresh_PF_DT;	/* K */
	float	thresh_PF_R2;	/* near-IR reflectance threshold, daytime algorithm only */

	/* thresholds used to identify valid bkg. pixels */
	float	thresh_bkg_T21;	/* K */
	float	thresh_bkg_DT;	/* K */

	float	thresh_T21;	/* K */

	float	minbkg_DT;	/* K; minimum difference above bkg. DT */

	float	devrp_T31;	/* K; T31 absolute deviation reduction parameter (daytime only) */
	} AlgParams;


static int test_potential_fire_pixel(AlgParams *ap, int i, int j,
	int first_line, Scan *PrevScan, Scan *scan, Scan *NextScan,
	ScanThreshPF *pfthresh, uint8 *FireMask, uint8 *ExtFireMask,
	uint32 *QA, FireCounts *counts);

static int BkgPixels(int ic, int jc, int surf,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT21, float *BuffT31, float *BuffDT, float *BuffL21,
	float *BuffR1, float *BuffR2, float *BuffR7, int *nref,
	float *BkgFireBuffT21, float *BkgFireBuffT31,
	float *BkgFireBuffDT, int *nbfire, int *winparam,
	int *nwater, int *nland, int *ncoast);

static void GetValidBkgPixels(int ic, int jc, int r, int surf_state,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT21, float *BuffT31, float *BuffDT, float *BuffL21,
	int *nvalid,
	float *BuffR1, float *BuffR2, float *BuffR7, int *nref,
	float *BkgFireBuffT21, float *BkgFireBuffT31, float *BkgFireBuffDT,
	int *nbfire, int *nwater, int *nland, int *ncoast);

static void CountAdjacentPixels(int ic, int jc,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	int *AdjCloud, int *AdjWater);

static void compute_stats(float *x, int n, float *mean, float *mad);
static int SunGlint(float, float, float, float, float, float, float *);
static void compute_ndvi(float *R1, float *R2, int n, float *NDVI);
static int bkg_water_pix(float *R2, float *R7, float *NDVI, int n);

static float fixang(float);

static float calc_thresh_PF_T21(float zone_mean_T21);
static float calc_thresh_PF_DT(float zone_mean_DT);



static AlgParams DayParams = {
	310.0F,	10.0F,	/* no fire */
	0.35F,		/* max. band 2 reflectance for fire pixels */

	325.0F,	20.0F,	/* valid bkg. */

	360.0F,		/* absolute threshold test */

	6.0F,

	4.0F
	};


static AlgParams NightParams = {
	305.0F,	10.0F,	/* no fire */
	0.0F,		/* dummy band 2 "nighttime reflectance" */

	310.0F,	10.0F,	/* valid bkg. */

	320.0F,		/* absolute threshold test */

	6.0F,

	1000.0F		/* dummy nighttime value that will cause test to always be passed */
	};





extern UBYTE iprev[SAMPLES_1KM], inext[SAMPLES_1KM];



void prescreen(Scan *scan, int skip_water)
/*
!C*****************************************************************************
!Description:
	Pre-screen current buffer prior to performing detection algorithm.
	Sets individual flags within the "pixinfo" field of the Scan
	structure pointed to by "scan".
	This function will also swap contents of individual elements of
	scan->T21 and scan->T22 arrays as necessary to obtain valid,
	unsaturated mid-infrared brightness temperatures to process.

!Input Parameters:
	skip_water	flag: exclude all water pixels from processing
			(For the C6 algorithm this flag is normally clear.
			DB ground stations may wish to skip water pixels
			when processing, however, so this flag is
			included in the code.)
!Output Parameters:

!Input/Output Parameters:
	scan		pointer to current scan buffer

!Returns:
	Nothing.

!Revision History

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;
	int surf;
	int processed_water;	/* flag: water pixel that will be processed */
	int process;		/* flag: land or processed-water pixel */
	int day;		/* flag: daytime pixel */
	int bad_geo;		/* flag: invalid geolocation for pixel */

	int glint;		/* flag: glint = non-zero */
	float GlintAng;		/* degrees */

	int zflag;		/* flag: include pixel when computing
				zone statistics */

	float T21, T22, T31, T32;
	float L21, L22;
	int use22;		/* flag: use band 22 in place of 21 */
	float RelAzAng;		/* degrees */
	AlgParams *ap;

	const float MaxT22 = 329.0F;	/* use T21 if T22 exceeds this
					threshold */

	PixInfo NullPixInfo;


	(void) memset(&NullPixInfo, 0, sizeof(PixInfo));

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			/* clear internal flags */
			scan->pixflags[i][j] = NullPixInfo;

			T21 = scan->T21[i][j];
			T22 = scan->T22[i][j];
			T31 = scan->T31[i][j];
			T32 = scan->T32[i][j];

			L21 = scan->L21[i][j];
			L22 = scan->L22[i][j];


			/*
			Use band 22 if available and not saturated,
			otherwise use band 21.  Here we'll actually
			replace the band 21 brightness temperature
			with band 22 provided that either 1) band 22 is
			available and not saturated, or 2) band 21 is
			unavailable.
			*/
			use22 = (T22 > 0.0F) &&
				((T22 < MaxT22) || (T21 <= 0.0F));

			if (use22) {
				T21 = scan->T21[i][j] = T22;
				L21 = scan->L21[i][j] = L22;
				}
			else {
				T22 = scan->T22[i][j] = T21;
				L22 = scan->L22[i][j] = L21;

				/* set internal flag */
				scan->pixflags[i][j].high_range = 1;
				}


			/* check that we have valid geolocation */
			bad_geo = ((fabsf(scan->lat[i][j]) > 90.0) ||
				(fabsf(scan->lon[i][j]) > 180.0));
			if (bad_geo) {
				scan->pixflags[i][j].missing_geo = 1;
				continue;
				}

			/* set day/night flag */
			day = scan->pixflags[i][j].day =
				scan->SolZenAng[i][j] < 85.0F;

			/* treat MOD03 "land" and "ephemeral water" as land */
			if ((scan->landsea[i][j] == 1) ||
				(scan->landsea[i][j] == 4)) {
				surf = scan->pixflags[i][j].surf = LAND;
				processed_water = 0;
				}
			else {
				/* water or coast */

				/* "Ocean Coastlines and Lake Shorelines" = 2 */
				if (scan->landsea[i][j] == 2)
					surf = COAST;
				else
					surf = WATER;

				scan->pixflags[i][j].surf = surf;

				if (skip_water)
					processed_water = 0;
				else {
					/*
					Omit coast, "Shallow Inland Water" (3),
					an "Deep Inland Water" (5) from the
					fire detection algorithm.
					*/
					processed_water = (surf == WATER) &&
						(scan->landsea[i][j] != 3) &&
						(scan->landsea[i][j] != 5);
					}
				}


			/*
			Check availability of bands that are essential for fire
			detection.  (Note that at this point merging of bands
			21 and 22 has already been performed.)  We do this
			AFTER setting the day/night and land/water flags so
			that the CMG module can properly tally missing pixels.
			*/
			if ((T21 < 0.0F) || (T31 < 0.0F)) {
				scan->pixflags[i][j].missing_rad = 1;
				continue;
				}


			if (day) {
				/* sun-glint test */
				RelAzAng = fixang(scan->SolAzAng[i][j] -
					scan->ViewAzAng[i][j]);

				glint = SunGlint(scan->R1[i][j], scan->R2[i][j],
					scan->R7[i][j], scan->SolZenAng[i][j],
					scan->ViewZenAng[i][j],
					RelAzAng, &GlintAng);

				scan->pixflags[i][j].glint_level = glint;
				}
			else
				glint = 0;


			/* run cloud mask on all pixels */
			if (T32 > 0.0F) {
				if (iscloud(scan->R1[i][j], scan->R2[i][j],
					T32, surf == WATER)) {
					scan->pixflags[i][j].cloud = 1;
					continue;
					}
				}
/* else set missing cloud mask flag */

			/*
			At this point we have a cloud-free pixel.
			Detection algorithm will be applied to
			all land pixels, no coastal pixels, and only a subset
			of water pixels.
			*/

			scan->pixflags[i][j].process = process =
				(surf == LAND) || processed_water;


			/* stop processing current pixel if we're not going
			to apply the detection algorithm to it */
			if (!process) continue;

			/* at this point we have either a land pixel or
			a processed-water pixel */

			/* identify the subset of cloud-free pixels that are
			"safe" to use when computing zone statistics */
			if (day)
				ap = &DayParams;
			else
				ap = &NightParams;

			zflag = (T21 <= ap->thresh_T21) && (glint <= 1);

			scan->pixflags[i][j].zsafe = zflag;
			}
		}
}




/*
Calculate MODIS band 21 potential fire threshold given the mean T21 value
for a zone.
*/
static float calc_thresh_PF_T21(float zone_mean_T21)
{
	float thresh_T21;
	const float min_thresh_T21 = 300.0F;
	const float max_thresh_T21 = 330.0F;

	thresh_T21 = zone_mean_T21 + 5.0F;

	if (thresh_T21 < min_thresh_T21) return min_thresh_T21;
	if (thresh_T21 > max_thresh_T21) return max_thresh_T21;

	return thresh_T21;
}



static float calc_thresh_PF_DT(float zone_mean_DT)
{
	float thresh_DT;
	const float min_thresh_DT = 10.0F;
	const float max_thresh_DT = 35.0F;

	thresh_DT = zone_mean_DT + 5.0F;

	if (thresh_DT < min_thresh_DT) return min_thresh_DT;
	if (thresh_DT > max_thresh_DT) return max_thresh_DT;

	return thresh_DT;
}




/* compute potential fire thresholds for each sample in specified scan */
void compute_scan_PF_thresholds(Scan *scan, ScanThreshPF *pfthresh)
{
	int i, j;
	AlgParams *ap;

	float thresh_PF_T21, thresh_PF_DT;	/* kelvins */
	const float missval = -1.0F;	/* use negative value for missing */

	static float land_thresh_PF_T21[SAMPLES_1KM];
	static float land_thresh_PF_DT[SAMPLES_1KM];


	for (j = 0; j < SAMPLES_1KM; j++) {
		if (scan->LandZonePix[j] >= MIN_VALID_ZONE_PIX) {
			land_thresh_PF_T21[j] =
				calc_thresh_PF_T21(scan->LandMeanT21[j]);
			land_thresh_PF_DT[j] =
				calc_thresh_PF_DT(scan->LandMeanDT[j]);
			}
		else {
			land_thresh_PF_T21[j] = missval;
			land_thresh_PF_DT[j] = missval;
			}
		}


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			if (!scan->pixflags[i][j].process) {
				pfthresh->thresh_PF_T21[i][j] = missval;
				pfthresh->thresh_PF_DT[i][j] = missval;
				continue;
				}

			if (scan->pixflags[i][j].day)
				ap = &DayParams;
			else
				ap = &NightParams;

			if (scan->pixflags[i][j].surf == LAND) {
				thresh_PF_T21 = land_thresh_PF_T21[j];
				thresh_PF_DT = land_thresh_PF_DT[j];
				}
			else {
				/* force use of default thresholds over water */
				thresh_PF_T21 = thresh_PF_DT = missval;
				}

			/*
			Use conservative fixed potential-fire thresholds if
			sample statistics are unavailable, or if we are
			processing a water pixel.
			*/
			if (thresh_PF_T21 < 0.0F) {
				thresh_PF_T21 = ap->thresh_PF_T21;
				thresh_PF_DT = ap->thresh_PF_DT;
				}

			pfthresh->thresh_PF_T21[i][j] = thresh_PF_T21;
			pfthresh->thresh_PF_DT[i][j] = thresh_PF_DT;
			}
		}
}



/*
Flag potential fire pixels in specified scan.

Functions prescreen(), compute_zone_stats(), and compute_scan_PF_thresholds()
must be called prior to calling this function.
*/
void flag_potential_fires(Scan *scan, ScanThreshPF *pfthresh)
{
	int i, j;
	float T21, T31, DT;			/* kelvins */
	AlgParams *ap;
	int pftest1, pftest2, pftest3;

	float thresh_PF_T21, thresh_PF_DT;	/* kelvins */


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			if (!scan->pixflags[i][j].process) continue;

			T21 = scan->T21[i][j];
			T31 = scan->T31[i][j];
			DT = T21 - T31;

			if (scan->pixflags[i][j].day)
				ap = &DayParams;
			else
				ap = &NightParams;

			thresh_PF_T21 = pfthresh->thresh_PF_T21[i][j];
			thresh_PF_DT = pfthresh->thresh_PF_DT[i][j];

			pftest1 = T21 > thresh_PF_T21;
			pftest2 = DT > thresh_PF_DT;
			pftest3 = scan->R2[i][j] < ap->thresh_PF_R2;

			/* potential fire test */
			if (pftest1 && pftest2 && pftest3)
				scan->pixflags[i][j].potential_fire = 1;
			}
		}
}



/*
Flag background fire pixels in specified scan.
See 2003 RSE paper for definitions and details.

Function prescreen() must be called prior to calling this function.
*/
void flag_background_fires(Scan *scan)
{
	int i, j;
	float T21, T31, DT;			/* kelvins */
	AlgParams *ap;

	int bfire;	/* flag: background fire pixel */


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			if (!scan->pixflags[i][j].process) continue;

			T21 = scan->T21[i][j];
			T31 = scan->T31[i][j];
			DT = T21 - T31;

			if (scan->pixflags[i][j].day)
				ap = &DayParams;
			else
				ap = &NightParams;

			/* valid background pixel test */
			bfire = (T21 > ap->thresh_bkg_T21) &&
				(DT > ap->thresh_bkg_DT);

			scan->pixflags[i][j].bkg_fire = bfire;
			scan->pixflags[i][j].valid_bkg_pix = !bfire;
			}
		}
}





int detect(Scan *PrevScan, Scan *scan, Scan *NextScan, int first_line,
	int skip_water, ScanThreshPF *pfthresh, OutScan *outscan,
	FireCounts *counts)
/*
!C*****************************************************************************
!Description:
	Apply detection algorithm to current scan.

!Input Parameters:
	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

	first_line
	skip_water
	pfthresh

!Output Parameters:
	outscan
	counts

!Input/Output Parameters:
	None.

!Returns:
	0 if no errors, non-zero if fire pixel table functions encountered
	error allocating memory.

!Revision History
	Original version April 2001, L. Giglio.


!END***************************************************************************
*/
{
	int i, j;
	int day;		/* flag: day = non-zero, night = 0 */
	int land;
	int coast;
	AlgParams *ap;

	uint8 fmask, efmask;
	uint32 QA;

	int st;


	/* initialize the "inext" and "iprev" arrays for each scan */
	init_bowtie_tables(PrevScan, scan, NextScan);


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			/* clear all QA bits */
			outscan->FireQA[i][j] = 0U;
outscan->ExtFireMask[i][j] = ECLASS_BOGUS;

			day = scan->pixflags[i][j].day;
			land = scan->pixflags[i][j].surf == LAND;
			coast = scan->pixflags[i][j].surf == COAST;

			/* copy land/water state into QA bits 0 and 1 */
			outscan->FireQA[i][j] |= scan->pixflags[i][j].surf;

			if (day) {
				ap = &DayParams;

				/* set day/night algorithm flag in product */
				outscan->FireQA[i][j] |= BIT_MASK(4);

				counts->NumDayPix++;
				}
			else {
				ap = &NightParams;

				counts->NumNightPix++;
				}

			/* missing data (either radiance or geolocation)? */
			if (scan->pixflags[i][j].missing_rad ||
				scan->pixflags[i][j].missing_geo) {
				outscan->FireMask[i][j] = CLASS_NOPROC_MISS;
outscan->ExtFireMask[i][j] = ECLASS_MISS;
				counts->NumMissingPix++;
				counts->NumMissingRad +=
					scan->pixflags[i][j].missing_rad;
				counts->NumMissingGeo +=
					scan->pixflags[i][j].missing_geo;
				continue;
				}


			if ( !scan->pixflags[i][j].high_range )
				outscan->FireQA[i][j] |= BIT_MASK(2);

			/* update counts of glint pixels and set glint level */
			if (scan->pixflags[i][j].glint_level) {
				counts->NumGlintPix++;
				/* bits 22-23 */
				outscan->FireQA[i][j] |=
					(scan->pixflags[i][j].glint_level << 22);
				}


			if (land)
				counts->NumLandPix++;
			else {
				if (coast)
					counts->NumCoastPix++;
				else
					counts->NumWaterPix++;
				}


			/* mimic C5 handling of water pixels if requested */
			if (skip_water && !land) {
				outscan->FireMask[i][j] = CLASS_WATER;
				outscan->ExtFireMask[i][j] =
					ECLASS_UNKNOWN_WATER;
				continue;
				}


			/* cloud? */
			if (scan->pixflags[i][j].cloud) {
				if (land) {
					/* cloud over land */
					outscan->FireMask[i][j] = CLASS_CLOUD;
					outscan->ExtFireMask[i][j] =
						ECLASS_CLOUDY_LAND;
					counts->NumLandCloudPix++;
					}
				else {
					/* cloud over water or coast */
					outscan->FireMask[i][j] = CLASS_CLOUD;
					outscan->ExtFireMask[i][j] =
						ECLASS_CLOUDY_WATER;
					counts->NumWaterCloudPix++;
					}

				continue;
				}


			/* handle unprocessed clear-water pixels */
			if (!scan->pixflags[i][j].process) {
				/* coast or unprocessed water? */
				if (!land) {
					outscan->FireMask[i][j] = CLASS_WATER;
					outscan->ExtFireMask[i][j] =
						ECLASS_UNKNOWN_WATER;
					counts->NumUnknownWaterPix++;
					}
#ifdef DEBUG
				else {
					/* unprocessed land pixels should
					never have made it to this point */
puts("UNEXPECTED UNPROCESSED LAND PIXEL");
					}
#endif

				continue;
				}


			/* obvious non-fire? (== not a potential fire) */
			if ( !scan->pixflags[i][j].potential_fire ) {
				if (land) {
					outscan->FireMask[i][j] = CLASS_NO_FIRE;
					outscan->ExtFireMask[i][j] =
						ECLASS_CLEAR_LAND;
					}
				else {
					outscan->FireMask[i][j] = CLASS_WATER;
					outscan->ExtFireMask[i][j] =
						ECLASS_CLEAR_WATER;
					}

				counts->NumNoFirePix++;
				continue;
				}


			/* At this point we have a potential fire pixel,
			either over water or land. */

			/* set algorithm QA: potential fire pixel */
			outscan->FireQA[i][j] |= BIT_MASK(5);

			st = test_potential_fire_pixel(ap, i, j, first_line,
				PrevScan, scan, NextScan, pfthresh,
				&fmask, &efmask, &QA, counts);
			if (st) return st;

			outscan->FireMask[i][j] = fmask;
			outscan->ExtFireMask[i][j] = efmask;
			outscan->FireQA[i][j] |= QA;
			}
		}

	return 0;
}






/*
Apply fire detection tests to a potential fire pixel.


Input Variables
	ap
	i
	j
	PrevScan
	scan
	NextScan

Input/Output Variables
	outscan		pointer to output scan buffer; element [i,j] of
			FireMask, ExtFireMask, and FireQA fields will be
			set by this function

	counts		pointer to FireCounts structure; fields will be
			updated as necessary

Output Variables
	FireMask
	ExtFireMask
	QA


!Returns:
	0 if no errors, non-zero if fire pixel table functions encountered
	error allocating memory.

*/
static int test_potential_fire_pixel(AlgParams *ap, int i, int j,
	int first_line, Scan *PrevScan, Scan *scan, Scan *NextScan,
	ScanThreshPF *pfthresh,
	uint8 *FireMask, uint8 *ExtFireMask, uint32 *QA, FireCounts *counts)
{
	float T21, T31;
	float DT;
	int day;		/* flag: day = non-zero, night = 0 */
	int land;

	int test1, test2, test3, test4, test5, test6;
	int fire;		/* flag: fire pixel */

	int winparam;

	float MeanT21, MeanDT, MeanT31;
	float MadT21, MadDT, MadT31;
	float z21, z31;
	const float eps = 1.0e-10F;
	float MeanR2, MadR2;

	int AdjCloud, AdjWater;
	int C;			/* detection confidence */

	float MeanL21;
	float power;		/* FRP [MW] */
	float fdummy;

	float RelAzAng;		/* degrees */

	int st;

	int nvalid;
	float BuffT21[MAX_WIN_PIX], BuffT31[MAX_WIN_PIX], BuffDT[MAX_WIN_PIX];
	float BuffL21[MAX_WIN_PIX];
	int nref;
	float BuffR1[MAX_WIN_PIX], BuffR2[MAX_WIN_PIX], BuffR7[MAX_WIN_PIX];
	float BuffNDVI[MAX_WIN_PIX];

	/* pixels that appear to be unmasked water pixels contaminating
	the background sample over land */
	int nwater_cont;

	float BkgFireMeanT21, BkgFireMeanDT, BkgFireMeanT31;
	float BkgFireMadT21, BkgFireMadDT, BkgFireMadT31;

	int nbfire;
	float fbfire;	/* fraction of pixels w/in window that are background
			fire pixels */

	int nwater, nland, ncoast;

	int nwc; /* nwc = nwater + ncoast; equivalent to nwater in C5 code */


	float BkgFireBuffT21[MAX_WIN_PIX], BkgFireBuffT31[MAX_WIN_PIX],
		BkgFireBuffDT[MAX_WIN_PIX];

	int cmg_row, cmg_col;


	*QA = 0;


	T21 = scan->T21[i][j];
	T31 = scan->T31[i][j];
	DT = T21 - T31;
	day = scan->pixflags[i][j].day;
	land = scan->pixflags[i][j].surf == LAND;


	/* compute background statistics */
	nvalid = BkgPixels(i, j, scan->pixflags[i][j].surf,
		PrevScan, scan, NextScan,
		BuffT21, BuffT31, BuffDT, BuffL21,
		BuffR1, BuffR2, BuffR7, &nref,
		BkgFireBuffT21, BkgFireBuffT31, BkgFireBuffDT,
		&nbfire, &winparam, &nwater, &nland, &ncoast);


	if (nbfire > 0) {
		compute_stats(BkgFireBuffT21, nbfire,
			&BkgFireMeanT21, &BkgFireMadT21);
		compute_stats(BkgFireBuffDT, nbfire,
			&BkgFireMeanDT, &BkgFireMadDT);
		compute_stats(BkgFireBuffT31, nbfire,
			&BkgFireMeanT31, &BkgFireMadT31);
		}
	else {
		BkgFireMeanT21 = BkgFireMeanT31 = BkgFireMeanDT = 0.0F;
		/* this will force test 6 to fail */
		BkgFireMadT21 = BkgFireMadT31 = BkgFireMadDT = 0.0F;
		}


	if (nvalid > 0) {
		compute_stats(BuffT21, nvalid, &MeanT21, &MadT21);
		compute_stats(BuffDT, nvalid, &MeanDT, &MadDT);
		compute_stats(BuffT31, nvalid, &MeanT31, &MadT31);

		fbfire = (float) nbfire / (float) nvalid;

		z21 = (T21 - MeanT21) / (MadT21 + eps);
		z31 = (T31 - MeanT31) / (MadT31 + eps);

		test2 = DT > (MeanDT + 3.5F * MadDT);
		test3 = DT > (MeanDT + ap->minbkg_DT);
		test4 = T21 > (MeanT21 + 3.0F * MadT21);
		test5 = T31 > (MeanT31 + MadT31 - ap->devrp_T31);
		test6 = BkgFireMadT21 > 5.0F;
		}
	else {
		MeanT21 = MeanT31 = MeanDT = 0.0F;
		MadT21 = MadT31 = MadDT = 0.0F;
		z21 = z31 = 0.0F;
		test2 = test3 = test4 = test5 = test6 = 0;
		fbfire = 0.0F;
		}

	test1 = T21 > ap->thresh_T21;

	/* set background characterization field (bits 7-8) */
	*QA |= (winparam << 7);

	/* save individual test results */
	if (test1) *QA |= BIT_MASK(11);
	if (test2) *QA |= BIT_MASK(12);
	if (test3) *QA |= BIT_MASK(13);
	if (test4) *QA |= BIT_MASK(14);
	if (test5) *QA |= BIT_MASK(15);
	if (test6) *QA |= BIT_MASK(16);

	/*
	There are so many routes below by which a non-fire
	class can be assigned that we will assume this result
	by default and later modify if necessary.
	*/
	if (land) {
		*FireMask = CLASS_NO_FIRE;
		*ExtFireMask = ECLASS_CLEAR_LAND;
		}
	else {
		*FireMask = CLASS_WATER;
		*ExtFireMask = ECLASS_CLEAR_WATER;
		}


	if (day)
		fire = test1 || (test2 && test3 && test4 && (test5 || test6));
	else
		fire = test1 || (test2 && test3 && test4);


	if (!fire) {
		/* we have a non-fire pixel */

		if (land) {
			/* distinguish between "no fire" and "unknown" */
			if (nvalid > 0)
				counts->NumNoFirePix++;
			else {
				*FireMask = CLASS_UNKNOWN;
				*ExtFireMask = ECLASS_UNKNOWN_LAND;
				counts->NumUnknownLandPix++;
				}
			}
		else {
			/* distinguish between "no fire" and "unknown" */
			if (nvalid > 0)
				counts->NumNoFirePix++;
			else {
				*FireMask = CLASS_UNKNOWN;
				*ExtFireMask = ECLASS_UNKNOWN_WATER;
				counts->NumUnknownWaterPix++;
				}
			}

		return 0;
		}


	/* ----- we have a fire pixel ----- */

	if (land && nref) {
		compute_ndvi(BuffR1, BuffR2, nref, BuffNDVI);
		nwater_cont = bkg_water_pix(BuffR2, BuffR7,
			BuffNDVI, nref);
		compute_stats(BuffR2, nref, &MeanR2, &MadR2);
		}
	else {
		nwater_cont = 0;
		MeanR2 = MadR2 = -1.0F;
		}


	CountAdjacentPixels(i, j, PrevScan, scan, NextScan,
		&AdjCloud, &AdjWater);


	/* ----- rejection tests ----- */

	/* override detection if sun glint is present */
	nwc = nwater + ncoast;
	if ( (scan->pixflags[i][j].glint_level > 1) ||
		((AdjWater || nwc) && scan->pixflags[i][j].glint_level) ) {
		*QA |= BIT_MASK(24);
		counts->NumGlintRejectPix++;
		counts->NumNoFirePix++;
		return 0;
		}


	if (land) {
		/* ----- land-specific rejection tests ----- */

		/* override land detection if it's likely that this is a
		false alarm caused by excessive rejection of legitimate
		background pixels */
		if (day && !test1 && (fbfire > 0.1F) && (nbfire > 3) &&
			(BkgFireMeanT21 < 345.0F) &&
			(BkgFireMadT21 < 3.0F) && (scan->R2[i][j] > 0.15F) &&
			(T21 < (BkgFireMeanT21 + 6.0F*BkgFireMadT21))) {
			*QA |= BIT_MASK(25);
			counts->NumHotRejectPix++;
			counts->NumNoFirePix++;
			return 0;
			}

		/* override land fire pixels that are most likely to be
		false alarms caused by water-pixel contamination
		of the background */
		if (day && !test1 && nwater_cont) {
			*QA |= BIT_MASK(26);
			counts->NumCoastRejectLandPix++;
			counts->NumNoFirePix++;
			return 0;
			}

		/* Amazon forest-clearing rejection test */
		if (day && nref && (MeanR2 > 0.28F) && (T21 < 325.0F)
			&& (z31 > 3.7F)) {
			*QA |= BIT_MASK(27);
			counts->NumClearingRejectPix++;
			counts->NumNoFirePix++;
			return 0;
			}
		}
	else {
		/* ----- water-specific rejection tests ----- */

		/*
		Override water detection if it's likely that this is a
		false alarm caused by an unmasked land or coast pixel.
		The presence of any land or coast pixels in the background
		window is a good indication that this is a false alarm.
		*/
		if (!test1 && (nland || ncoast)) {
			*QA |= BIT_MASK(28);
			counts->NumCoastRejectWaterPix++;
			counts->NumNoFirePix++;
			return 0;
			}
		}

	/*
 	Fire pixels over water (gas flares) should not be penalized for
	having adjacent water pixels, so we zero AdjWater for water fire pixels.
	*/
	C = confidence(T21, z21, DT, MeanDT, MadDT, AdjCloud,
		AdjWater * land,
		day, pfthresh->thresh_PF_T21[i][j]);

	*FireMask = CLASS_FIRE_LOW + (C >= 30) + (C >= 80);
	counts->NumFirePix++;
	if (land) {
		*ExtFireMask = ECLASS_FIRE_LAND;
		counts->NumLandFirePix++;
		}
	else {
		*ExtFireMask = ECLASS_FIRE_WATER;
		counts->NumWaterFirePix++;
		}

	if (nvalid > 0) {
		/* compute mean L21 */
		compute_stats(BuffL21, nvalid, &MeanL21, &fdummy);

		power = fire_power(scan->L21[i][j], MeanL21, j);
		}
	else
		power = 0.0F;


	cmg_row_col((float) scan->lat[i][j], (float) scan->lon[i][j],
		&cmg_row, &cmg_col);

	RelAzAng = fixang(scan->SolAzAng[i][j] - scan->ViewAzAng[i][j]);

	st = append_fire_pixel(i + first_line, j,
		scan->lat[i][j], scan->lon[i][j],
		scan->R2[i][j], T21, T31, MeanT21, MeanT31, MeanDT,
		MadT21, MadT31, MadDT, AdjCloud, AdjWater,
		winparam, nvalid, C, power, land,
		MeanR2, MadR2, scan->ViewZenAng[i][j], scan->SolZenAng[i][j],
		RelAzAng, cmg_row, cmg_col);
	if (st) return st;

	if (AdjCloud) {
		*QA |= BIT_MASK(20);	/* was bit 21 in C5 */
		counts->NumCloudAdjFirePix++;
		}

	if (AdjWater) {
		*QA |= BIT_MASK(21);	/* was bit 22 in C5 */
		counts->NumWaterAdjFirePix++;
		}

	return 0;
}






static int BkgPixels(int ic, int jc, int surf,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT21, float *BuffT31, float *BuffDT, float *BuffL21,
	float *BuffR1, float *BuffR2, float *BuffR7, int *nref,
	float *BkgFireBuffT21, float *BkgFireBuffT31,
	float *BkgFireBuffDT, int *nbfire,
	int *winparam, int *nwater, int *nland, int *ncoast)
/*
!C*****************************************************************************
!Description:
	Fill buffers with band 21 and 31 brightness temperatures (T21 and T31),
	and their difference (DT), for valid background pixels contained in the
	smallest square background window needed to characterize the background
	of a potential fire pixel.

	Window starts as 3 x 3 pixels in size, always excluding the two
	along-scan pixels adjacent to the candidate fire pixel, and is expanded
	as necessary until either a sufficient number of background pixels have
	been acquired, or the window becomes 21 x 21 pixels in size a
	sufficient number of background pixels still has not been obtained.  In
	this case, the characterization is deemed unsuccessful.

!Input Parameters:
	ic		scan position of potential fire pixel
	jc		sample position of potential fire pixel
	PrevScan
	CurrScan
	NextScan

!Output Parameters:
	BuffT21		buffer containing T21 for valid background pixels
	BuffT31		buffer containing T31 for valid background pixels
	BuffDT		buffer containing DT for valid background pixels
	BuffL21		buffer containing L21 ""

	BuffR1
	BuffR2
	BuffR7
	
	nref

	BkgFireBuffT21
	BkgFireBuffT31
	BkgFireBuffDT

	nbfire

	winparam	size parameter for window needed to characterize
			background

	nwater
	nland
	ncoast

!Input/Output Parameters:

!Returns:
	Number of valid background pixels, or 0 if an insufficient number of
	valid background pixels could be obtained.

!Revision History
	Original version 1997, L. Giglio.
	May 2002 updated code to reflect version 4 algorithm changes.
	Dec 2008 code clean-up.

!Team-unique Header:

!References and Credits

!Design Notes
	Window size is related to size parameter as:

		window size in pixels = 2 * winparam + 1

!END***************************************************************************
*/
{
	int nvalid, minvalid;
	int winsize;
	int r;
	const int rmax = 10;	/* corresponds to 21 by 21 window */


	nvalid = *nref = *nbfire = *nwater = *nland = *ncoast = 0;
	r = 1;		/* start at 3 x 3 ring */
	while (r <= rmax) {
		*winparam = r;
		winsize = 2*r + 1;

		GetValidBkgPixels(ic, jc, r, surf, PrevScan, CurrScan, NextScan,
			BuffT21, BuffT31, BuffDT, BuffL21, &nvalid,
			BuffR1, BuffR2, BuffR7, nref,
			BkgFireBuffT21, BkgFireBuffT31, BkgFireBuffDT, nbfire,
			nwater, nland, ncoast);

		minvalid = (int) (0.25F * (winsize*winsize - 3));		/* 25% */
		if (minvalid < 8) minvalid = 8;

		if (nvalid >= minvalid) return nvalid;

		r++;
		}

	*winparam = *nref = *nbfire = *nwater = *nland = *ncoast = 0;
	return 0;
}



static void GetValidBkgPixels(int ic, int jc, int r, int surf_state,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT21, float *BuffT31, float *BuffDT, float *BuffL21,
	int *nvalid,
	float *BuffR1, float *BuffR2, float *BuffR7, int *nref,
	float *BkgFireBuffT21, float *BkgFireBuffT31, float *BkgFireBuffDT,
	int *nbfire, int *nwater, int *nland, int *ncoast)
/*
!C*****************************************************************************
!Description:
	Append ch. 21, ch. 31, and ch. 21 - ch. 31 brightness temperatures of
	valid background pixels from a square "ring" of pixels to buffers.

!Input Parameters:
	ic		scan position around which ring will be centered
	jc		sample position around which ring will be centered
	r		"radius" of square ring

	surf_state	surface type for which we are to accumulate pixels
			(land or water)

	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

!Output Parameters:
	BuffT21		buffer containing T21 (K) for valid background pixels
	BuffT31		buffer containing T31 (K) for valid background pixels
	BuffDT		buffer containing DT (K) for valid background pixels
	BuffL21		buffer containing L21 (W/m^2/str/micron) "

!Input/Output Parameters:
	nvalid		number of valid background pixels contained in buffers
	nref
	nbfire
	nwater		number of water pixels present in background window
	nland		number of land pixels present in background window
	ncoast		number of coastal pixels present in background window

!Returns:

!Revision History
	Original version 1997, L. Giglio.

	Modified to deal with MODIS scan bow tie overlap near scan edges,
	allowing contextual algorithm to be applied even way off nadir.
	L. Giglio, Feb. 2002.

	Modified to also tabulate values for all background fire pixels,
	Mar. 2002.

	Oct. 2010, L. Giglio.  Heavily modified to accomodate processing of
	water pixels.

!END***************************************************************************
*/
{
	int i, j;
	int jstart, jstop;
	int jl, jr;		/* left and right samples */
	Scan *scan;

	int icorr;		/* bow-tie-corrected line index */


	if ( (jstart = jc - r) < 0) jstart = 0;
	if ( (jstop = jc + r) >= SAMPLES_1KM) jstop = SAMPLES_1KM - 1;



	/* top edge */
	i = ic - r;
	if (i < 0) {
		/* look in previous scan, handling bow tie */

		for (j = jstart; j <= jstop; j++) {
			icorr = i + 1 + iprev[j];
			if (icorr < 0) continue;

			if (PrevScan->pixflags[icorr][j].surf == LAND)
				(*nland)++;
			else {
				if (PrevScan->pixflags[icorr][j].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (PrevScan->pixflags[icorr][j].surf != surf_state)
				continue;

			if (PrevScan->pixflags[icorr][j].valid_bkg_pix) {
				BuffT21[*nvalid] = PrevScan->T21[icorr][j];
				BuffT31[*nvalid] = PrevScan->T31[icorr][j];
				BuffDT[*nvalid] = BuffT21[*nvalid] -
					BuffT31[*nvalid];
				BuffL21[*nvalid] = PrevScan->L21[icorr][j];
				(*nvalid)++;

				if (PrevScan->R2[icorr][j] > 0.0F) {
					BuffR1[*nref] = PrevScan->R1[icorr][j];
					BuffR2[*nref] = PrevScan->R2[icorr][j];
					BuffR7[*nref] = PrevScan->R7[icorr][j];
					(*nref)++;
					}
				}

			if (PrevScan->pixflags[icorr][j].bkg_fire) {
				BkgFireBuffT21[*nbfire] =
					PrevScan->T21[icorr][j];
				BkgFireBuffT31[*nbfire] =
					PrevScan->T31[icorr][j];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT21[*nbfire] -
					BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}
			}
		}
	else {
		/* no bow tie */
		for (j = jstart; j <= jstop; j++) {
			if (CurrScan->pixflags[i][j].surf == LAND)
				(*nland)++;
			else {
				if (CurrScan->pixflags[i][j].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (CurrScan->pixflags[i][j].surf != surf_state)
				continue;

			if (CurrScan->pixflags[i][j].valid_bkg_pix) {
				BuffT21[*nvalid] = CurrScan->T21[i][j];
				BuffT31[*nvalid] = CurrScan->T31[i][j];
				BuffDT[*nvalid] = BuffT21[*nvalid] -
					BuffT31[*nvalid];
				BuffL21[*nvalid] = CurrScan->L21[i][j];
				(*nvalid)++;

				if (CurrScan->R2[i][j] > 0.0F) {
					BuffR1[*nref] = CurrScan->R1[i][j];
					BuffR2[*nref] = CurrScan->R2[i][j];
					BuffR7[*nref] = CurrScan->R7[i][j];
					(*nref)++;
					}
				}

			if (CurrScan->pixflags[i][j].bkg_fire) {
				BkgFireBuffT21[*nbfire] = CurrScan->T21[i][j];
				BkgFireBuffT31[*nbfire] = CurrScan->T31[i][j];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT21[*nbfire] -
					BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}
			}
		}




	/* bottom edge */
	i = ic + r;
	if (i >= LINES_PER_SCAN_1KM) {
		/* look in next scan, handling bow tie */

		for (j = jstart; j <= jstop; j++) {
			icorr = i - LINES_PER_SCAN_1KM + inext[j];
			if (icorr >= LINES_PER_SCAN_1KM) continue;

			if (NextScan->pixflags[icorr][j].surf == LAND)
				(*nland)++;
			else {
				if (NextScan->pixflags[icorr][j].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (NextScan->pixflags[icorr][j].surf != surf_state)
				continue;

			if (NextScan->pixflags[icorr][j].valid_bkg_pix) {
				BuffT21[*nvalid] = NextScan->T21[icorr][j];
				BuffT31[*nvalid] = NextScan->T31[icorr][j];
				BuffDT[*nvalid] = BuffT21[*nvalid] -
					BuffT31[*nvalid];
				BuffL21[*nvalid] = NextScan->L21[icorr][j];
				(*nvalid)++;

				if (NextScan->R2[icorr][j] > 0.0F) {
					BuffR1[*nref] = NextScan->R1[icorr][j];
					BuffR2[*nref] = NextScan->R2[icorr][j];
					BuffR7[*nref] = NextScan->R7[icorr][j];
					(*nref)++;
					}
				}

			if (NextScan->pixflags[icorr][j].bkg_fire) {
				BkgFireBuffT21[*nbfire] =
					NextScan->T21[icorr][j];
				BkgFireBuffT31[*nbfire] =
					NextScan->T31[icorr][j];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT21[*nbfire] -
					BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}
			}
		}
	else {
		/* no bow tie */
		for (j = jstart; j <= jstop; j++) {
			if (CurrScan->pixflags[i][j].surf == LAND)
				(*nland)++;
			else {
				if (CurrScan->pixflags[i][j].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (CurrScan->pixflags[i][j].surf != surf_state)
				continue;

			if (CurrScan->pixflags[i][j].valid_bkg_pix) {
				BuffT21[*nvalid] = CurrScan->T21[i][j];
				BuffT31[*nvalid] = CurrScan->T31[i][j];
				BuffDT[*nvalid] = BuffT21[*nvalid] -
					BuffT31[*nvalid];
				BuffL21[*nvalid] = CurrScan->L21[i][j];
				(*nvalid)++;

				if (CurrScan->R2[i][j] > 0.0F) {
					BuffR1[*nref] = CurrScan->R1[i][j];
					BuffR2[*nref] = CurrScan->R2[i][j];
					BuffR7[*nref] = CurrScan->R7[i][j];
					(*nref)++;
					}
				}

			if (CurrScan->pixflags[i][j].bkg_fire) {
				BkgFireBuffT21[*nbfire] = CurrScan->T21[i][j];
				BkgFireBuffT31[*nbfire] = CurrScan->T31[i][j];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT21[*nbfire] -
					BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}
			}
		}



	/* if doing 3 x 3 window, i.e. r == 1, skip pixels adjacent to the
	candidate fire pixel in the along-scan direction */
	if (r == 1) return;



	jl = jc - r;	/* left edge */
	jr =  jc + r;	/* right edge */

	/*
	value of "i"			desired value of "scan"
	------------			-----------------------
	i < 0				scan = PrevPtr

	0 <= i < LINES_PER_SCAN_1KM	scan = CurrPtr

	i >= LINES_PER_SCAN_1KM		scan = NextPtr
	*/

	/* left side */
	if (jl >= 0) {
		for (i = ic - r + 1; i < ic + r; i++) {
			/* initialize as if pixel is in current scan buffer */
			scan = CurrScan;
			icorr = i;

			/* ...but reset if necessary */
			if (i < 0) {
				scan = PrevScan;
				icorr = i + 1 + iprev[jl];
				if (icorr < 0) continue;
				}
			else if (i >= LINES_PER_SCAN_1KM) {
				scan = NextScan;
				icorr = i - LINES_PER_SCAN_1KM + inext[jl];
				if (icorr >= LINES_PER_SCAN_1KM) continue;
				}

			if (scan->pixflags[icorr][jl].surf == LAND)
				(*nland)++;
			else {
				if (scan->pixflags[icorr][jl].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (scan->pixflags[icorr][jl].surf != surf_state)
				continue;

			if (scan->pixflags[icorr][jl].valid_bkg_pix) {
				BuffT21[*nvalid] = scan->T21[icorr][jl];
				BuffT31[*nvalid] = scan->T31[icorr][jl];
				BuffDT[*nvalid] = BuffT21[*nvalid] -
					BuffT31[*nvalid];
				BuffL21[*nvalid] = scan->L21[icorr][jl];
				(*nvalid)++;

				if (scan->R2[icorr][jl] > 0.0F) {
					BuffR1[*nref] = scan->R1[icorr][jl];
					BuffR2[*nref] = scan->R2[icorr][jl];
					BuffR7[*nref] = scan->R7[icorr][jl];
					(*nref)++;
					}
				}

			if (scan->pixflags[icorr][jl].bkg_fire) {
				BkgFireBuffT21[*nbfire] = scan->T21[icorr][jl];
				BkgFireBuffT31[*nbfire] = scan->T31[icorr][jl];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT21[*nbfire] -
					BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}
			}
		}


	/* right side */
	if (jr < SAMPLES_1KM) {
		for (i = ic - r + 1; i < ic + r; i++) {
			/* initialize as if pixel is in current scan buffer */
			scan = CurrScan;
			icorr = i;

			/* ...but reset if necessary */
			if (i < 0) {
				scan = PrevScan;
				icorr = i + 1 + iprev[jr];
				if (icorr < 0) continue;
				}
			else if (i >= LINES_PER_SCAN_1KM) {
				scan = NextScan;
				icorr = i - LINES_PER_SCAN_1KM + inext[jr];
				if (icorr >= LINES_PER_SCAN_1KM) continue;
				}

			if (scan->pixflags[icorr][jr].surf == LAND)
				(*nland)++;
			else {
				if (scan->pixflags[icorr][jr].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (scan->pixflags[icorr][jr].surf != surf_state)
				continue;

			if (scan->pixflags[icorr][jr].valid_bkg_pix) {
				BuffT21[*nvalid] = scan->T21[icorr][jr];
				BuffT31[*nvalid] = scan->T31[icorr][jr];
				BuffDT[*nvalid] = BuffT21[*nvalid] -
					BuffT31[*nvalid];
				BuffL21[*nvalid] = scan->L21[icorr][jr];
				(*nvalid)++;

				if (scan->R2[icorr][jr] > 0.0F) {
					BuffR1[*nref] = scan->R1[icorr][jr];
					BuffR2[*nref] = scan->R2[icorr][jr];
					BuffR7[*nref] = scan->R7[icorr][jr];
					(*nref)++;
					}
				}

			if (scan->pixflags[icorr][jr].bkg_fire) {
				BkgFireBuffT21[*nbfire] = scan->T21[icorr][jr];
				BkgFireBuffT31[*nbfire] = scan->T31[icorr][jr];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT21[*nbfire] -
					BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}
			}
		}
}



static void CountAdjacentPixels(int ic, int jc,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	int *AdjCloud, int *AdjWater)
/*
!C*****************************************************************************
!Description:
	Count number of cloud and water pixels adjacent to detected
	fire pixel.  Here both coast and water pixels are considered to be
	water.

!Input Parameters:
	ic		scan position at which to count adjacent pixels
	jc		sample position at which to count adjacent pixels

	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

!Output Parameters:
	AdjCloud	number of adjacenct cloud pixels (range 0 - 8)
	AdjWater	number of adjacent water pixels (range 0 - 8)

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History
	Original version: L. Giglio, 1997.

	Corrected version: L. Giglio, July 2013.  Code did not properly handle
	bow tie and therefore did not count adjacent pixels correctly when the
	central pixel was located in the first or last line of the current
	scan.  This long standing bug (> 1 decade) was due to my carelessness
	in using the pre-launch version of this function (at which time the
	detection algorithm ignored the bow tie regions) in the Collection 4
	and Collection 5 codes.

!END***************************************************************************
*/
{
	int i, j;
	int jstart, jstop;
	int icorr;		/* bow-tie-corrected line index */

	*AdjCloud = *AdjWater = 0;


	/* pixels to the left and right of center pixel */
	if (jc > 0) {
		if (CurrScan->pixflags[ic][jc-1].surf != LAND) (*AdjWater)++;
		if (CurrScan->pixflags[ic][jc-1].cloud) (*AdjCloud)++;
		}

	if (jc < (SAMPLES_1KM - 1)) {
		if (CurrScan->pixflags[ic][jc+1].surf != LAND) (*AdjWater)++;
		if (CurrScan->pixflags[ic][jc+1].cloud) (*AdjCloud)++;
		}


	/* sets of three pixels above and below center pixel */
	if ( (jstart = jc - 1) < 0) jstart = 0;
	if ( (jstop = jc + 1) >= SAMPLES_1KM) jstop = SAMPLES_1KM - 1;

	/* top edge */
	i = ic - 1;
	if (i < 0) {
		/* look in previous scan, handling bow tie */
		for (j = jstart; j <= jstop; j++) {
			icorr = i + 1 + iprev[j];
			if (icorr < 0) continue;

			if (PrevScan->pixflags[icorr][j].surf != LAND)
				(*AdjWater)++;
			if (PrevScan->pixflags[icorr][j].cloud) (*AdjCloud)++;
			}
		}
	else {
		/* no bow tie */
		for (j = jstart; j <= jstop; j++) {
			if (CurrScan->pixflags[i][j].surf != LAND)
				(*AdjWater)++;
			if (CurrScan->pixflags[i][j].cloud) (*AdjCloud)++;
			}
		}


	/* bottom edge */
	i = ic + 1;
	if (i >= LINES_PER_SCAN_1KM) {
		/* look in next scan, handling bow tie */
		for (j = jstart; j <= jstop; j++) {
			icorr = i - LINES_PER_SCAN_1KM + inext[j];
			if (icorr >= LINES_PER_SCAN_1KM) continue;

			if (NextScan->pixflags[icorr][j].surf != LAND)
				(*AdjWater)++;
			if (NextScan->pixflags[icorr][j].cloud) (*AdjCloud)++;
			}
		}
	else {
		/* no bow tie */
		for (j = jstart; j <= jstop; j++) {
			if (CurrScan->pixflags[i][j].surf != LAND)
				(*AdjWater)++;
			if (CurrScan->pixflags[i][j].cloud) (*AdjCloud)++;
			}
		}
}




/*
Alg. will only call this function with n >= 1 so no checks for n == 0
are needed.
*/
static void compute_stats(float *x, int n, float *mean, float *mad)
{
	int	i;
	double	sum;

	/* mean */
	for (i = 0, sum = 0.0; i < n; i++) sum += (double) x[i];
	*mean = (float) (sum / n);

	/* mean absolute deviation */
	for (i = 0, sum = 0.0; i < n; i++)
		sum += (double) fabsf(x[i] - *mean);

	*mad = (float) (sum / n);
}




static int SunGlint(float R1, float R2, float R7, float ts, float tv, float phi,
	float *GlintAng)
/*
!C*****************************************************************************
!Description:
	Perform sun-glint test specified in Giglio et al. 2003.

!Input Parameters:
	R1	channel 1 (0.64 micron) 1 km aggregated reflectance
	R2	channel 2 (0.86 micron) 1 km aggregated reflectance
	R7	channel 7 (2.1 micron)  1 km aggregated reflectance
	ts	solar zenith angle ("theta s") in degrees
	tv	view zenith angle ("theta v") in degrees
	phi	relative azimuth angle in degrees

!Output Parameters:
	GlintAng	"glint" angle (degrees)

!Input/Output Parameters:
	None.

!Returns:
	0 if sun glint absent, otherwise glint level from 1 (lowest)
	to 3 (highest)

!Revision History
	Original version 1997, L. Giglio.
	Used mutated Li et al. tests.

	L. Giglio, 14 May 1998
	Glint angle increased to from 30 to 40 degrees as specified by Y.
	Kaufman.

	L. Giglio, 26 Feb. 2001
	R1 and R2 thresholds decreased from 0.30 to 0.28.

	L. Giglio, 28 Feb. 2002
	Changed glint angle calculation to match that in Giglio, et al. 2002.
	Old method used E. Prins calculation but this suffered from a sign
	error.

	August 2002
	Added band 7 and modified reflective channel thresholds.

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	float cos_tr;		/* cosine of reflected sun angle "theta r" */
	float tr;

	cos_tr = cosf(D2R * tv) * cosf(D2R * ts) -
		sinf(D2R * tv) * sinf(D2R * ts) * cosf(D2R * phi);

	tr = R2D * acosf(cos_tr);

	*GlintAng = tr;

	/* automatically assume glint for very small glint angles */
	if (tr < 2.0F) return 3;

        if ((tr < 10.0F) && (R1 > 0.10F) && (R2 > 0.20F) && (R7 > 0.12F))
		return 2;

	if (tr < 15.0F) return 1;

	return 0;
}






/* does NOT handle missing data in input arrays; make sure there isn't any when calling */
static void compute_ndvi(float *R1, float *R2, int n, float *NDVI)
{
	int i;

	for (i = 0; i < n; i++)
		NDVI[i] = (R2[i] - R1[i]) / (R2[i] + R1[i]);
}



/* returns number of background pixels that are very likely to contain water */
static int bkg_water_pix(float *R2, float *R7, float *NDVI, int n)
{
	int i;
	int nwater;

	for (i = nwater = 0; i < n; i++)
		if ((R2[i] < 0.15F) && (R7[i] < 0.05F) && (NDVI[i] < 0.0F))
			nwater++;

	return nwater;
}


/* ===== utility functions ===== */


static float fixang(float ang)
/*
!C*****************************************************************************
!Description:
	Map angle in range [-540,+540] to range [-180,+180].

!Input Parameters:
	ang	angle (degrees)

!Output Parameters:
	None.

!Input/Output Parameters:
	None.

!Returns:
	Returns angle mapped to range [-180,+180]

!Revision History

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	if (ang > +180.0F) return ang - 360.0F;
	if (ang <= -180.0F) return 360.0F + ang;
	return ang;
}




