/*
!Description:

	Implementation of the Collection 6 MODIS active fire detection
	algorithm.  The earliest version of the algorithm, which is now
	obsolete, is described in:

	Kaufman, Y., Justice, C. O., Flynn, L. P., Kendall, J. D.,
	Prins, E. M, Giglio, L., Ward, D. E., Menzel, W. P., and
	Setzer, A. W., 1998, Potential global fire monitoring from
	EOS-MODIS.  Journal of Geophysical Research, 103:32215-32238.

	The Collection 4 and 5 algorithm is described in:

	Giglio, L., Descloitres, J., Justice, C. O., & Kaufman, Y. J. (2003).
	An enhanced contextual fire detection algorithm for MODIS.  Remote
	Sensing of Environment, 87:273-282.

	The Collection 6 algorithm adds modest improvements to the
	Collection 5 algorithm and will be described in a
	forthcoming paper.

	For additional information see the MODIS Active Fire Product and
	Burned Area Product web site (http://modis-fire.umd.edu/).

!References and Credits:

	Louis Giglio
	Wilfrid Schroeder
	Christopher O. Justice
	University of Maryland, Department of Geographical Sciences
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>


#include "mfhdf.h"


#include "const.h"
#include "types.h"
#include "tb.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "io.h"
#include "process.h"
#include "cmg.h"

#include "error.h"
#include "version.h"







int verbose;		/* flag: verbose mode */
int write_coarse;	/* flag: generate separate coarse-resolution product (MOD14CRS) */
int write_cmg;		/* flag: append CMG input-data layer to L2 product */
int write_tb;		/* flag: append brightness temp. SDSs to L2 product */
int write_ref;		/* flag: append reflectance SDSs to L2 product */
int write_geo;		/* flag: append geolocation SDSs to L2 product */
int coast_adjust;	/* flag: perform C6 MOD03/MYD03 coast adjustment */


int main(int argc, char **argv)
{
	int32 sd_mod02, sd_mod03;
	int32 sd_fire;
	int32 sd_coarse_fire;	/* MOD14CRS */

	char *infile_mod02;	/* 1 km 1B product */
	char *infile_mod03;

	HDF_InputInfo hin;
	HDF_OutputInfo hout;

	char *outfile, *outfile_coarse;
	int st;

	char msg[MAX_MSG_LEN];

	FireCounts counts;

	int skip_water;		/* flag: do not process water pixels */

	int c;
	extern char *optarg;
	extern int optind, opterr;


	write_coarse = 0;
	outfile_coarse = (char *) NULL;
	verbose = write_cmg = 0;
	write_tb = write_ref = write_geo = 0;
	skip_water = 0;

	/*
 	The C6 MOD03/MYD03 coastline adjustment is enabled by default.
	Note that this adjustment will be performed ONLY if the "WaterPresent"
	SDS is present in the MOD03/MYD03 input file, regardless of the state
	of this flag, to preserve backwards compatability with the pre-C6
	MODIS geolocation product.
	*/
	coast_adjust = -1;

	while ((c = getopt(argc, argv, "c:CgLrtvwy")) != -1) {
		switch (c) {
			case 'v':
				verbose = -1;
				break;

			case 'L':
				write_geo = -1;
				break;

			case 'r':
				write_ref = -1;
				break;

			case 't':
				write_tb = -1;
				break;

			case 'c':
				write_coarse = -1;
				outfile_coarse = optarg;
				break;

			case 'g':
				write_cmg = -1;
				break;

			case 'C':
				fputs("Obsolete -C switch will be ignored.\n",
					stderr);
				break;

			case 'w':
				skip_water = -1;
				break;

			case 'y':
				coast_adjust = 0;
				break;

			default:
				USAGE:
				fputs("Usage: mod14 [-gLrtvwy] [-c coarse_output] MOD021KM_input"
					" MOD03_input output_file\n", stderr);
				fputs("    c    Write specified coarse-resolution product file.\n", stderr);
				fputs("    g    Add CMG SDS to L2 product.\n", stderr);
				fputs("    L    Add geolocation SDSs to L2 product.\n", stderr);
				fputs("    r    Add R7 reflectance SDSs to L2 product.\n", stderr);
				fputs("    t    Add T21, T31, T32 brightness temperature SDSs to L2 product.\n", stderr);
				fputs("    v    Verbose mode.\n", stderr);
				fputs("    w    Skip water pixels during processing (a la C5 and earlier).\n", stderr);
				fputs("    y    Do not perform C6 coastline adjustment (C6 MOD03/MYD03 only).\n", stderr);
				fprintf(stderr, "\nVersion %s, compiled %s %s.\n",
					PROCESS_VERSION_NUMBER, __DATE__, __TIME__);
				exit(1);
				break;
			}
		}


	/* make sure three file names follow any switches */
	if (optind != argc - 3) goto USAGE;

	infile_mod02 = argv[optind];
	infile_mod03 = argv[optind+1];
	outfile = argv[optind+2];


	if (verbose) printf("%s\n%s\n%s\n", infile_mod02, infile_mod03, outfile);


	if ((sd_mod02 = SDstart(infile_mod02, DFACC_READ)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF input file \"%s\".", infile_mod02);
		errmsg(msg);
		exit(1);
		}

	if ((sd_mod03 = SDstart(infile_mod03, DFACC_READ)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF input file \"%s\".", infile_mod03);
		errmsg(msg);
		exit(2);
		}

	if ( (st = InitInputHDF(sd_mod02, sd_mod03, &hin)) ) exit(st);

	if (verbose) printf("nscans = %d\n", hin.nscans);

	if ((sd_fire = SDstart(outfile, DFACC_CREATE)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF output file \"%s\".", outfile);
		errmsg(msg);
		exit(3);
		}

	if (write_coarse) {
		if ((sd_coarse_fire = SDstart(outfile_coarse, DFACC_CREATE)) == FAIL) {
			(void) error(ERR_HDF_QUIET);
			(void) sprintf(msg, "Unable to open HDF output file \"%s\".", outfile);
			errmsg(msg);
			exit(3);
			}
		}
	else
		sd_coarse_fire = -1;

	st = InitOutputHDF(sd_fire, sd_coarse_fire, hin.nscans, write_geo,
		&hout);
	if (st) exit(4);


	if (cmg_init(write_cmg)) {
		exit(4);
		}

	st = ProcGranule(&hin, &hout, &counts, skip_water);
	if (st) {
		fputs("Error processing granule.\n", stderr);
		exit(3);
		}

	if (verbose) {
		printf("Fire:       %7d\n", counts.NumFirePix);
		printf("Unknown:    %7d\n", counts.NumUnknownLandPix);
		}

	if (write_cmg) {
if (verbose) puts("Writing CMG layers...");
		if (cmg_write(hout.sd_id_fire, CMG_DAY)) exit(5);
		if (cmg_write(hout.sd_id_fire, CMG_NIGHT)) exit(5);
		}

	if (write_generic_attributes(hout.sd_id_fire, infile_mod02,
		infile_mod03, hin.sat)) exit(5);


	if (write_coarse)
		if (SDend(sd_coarse_fire) == FAIL) exit(error(ERR_HDF));
	if (SDend(sd_fire) == FAIL) exit(error(ERR_HDF));
	if (SDend(sd_mod03) == FAIL) exit(error(ERR_HDF));
	if (SDend(sd_mod02) == FAIL) exit(error(ERR_HDF));

	free_fire_pixel_table();

	cmg_free(write_cmg);

	exit(0);
}


